from typing import Self

import xarray
from pathlib import Path

from torch.utils.data import Dataset
from xarray import Dataset as XarrayDataset


class XarrayZarrDataset(Dataset):
    @classmethod
    def new(cls, path: Path) -> Self:
        xarray_dataset: XarrayDataset = xarray.open_zarr(path)
        instance = cls(xarray_dataset=xarray_dataset)
        return instance

    def __init__(self, xarray_dataset: XarrayDataset):
        self.xarray_dataset: XarrayDataset = xarray_dataset

    def __len__(self):
        return self.xarray_dataset['index'].size

    def __getitem__(self, index):
        input_ = self.xarray_dataset['input'][index]
        output = self.xarray_dataset['output'][index]
        return input_, output



from haplo.internal.constantinos_kalapotharakos_format import \
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr
from haplo.internal.mcmc_output_analysis import mcmc_output_xarray_dataset_to_pandas_data_frame, \
    slice_iteration_of_mcmc_output_xarray_dataset

__all__ = [
    'combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr',
    'mcmc_output_xarray_dataset_to_pandas_data_frame',
    'slice_iteration_of_mcmc_output_xarray_dataset',
]




from __future__ import annotations

import math

import logging
import mmap
import numpy as np
import re
from pathlib import Path
from typing import TextIO, Dict, List

import pandas as pd
import polars as pl

from haplo.logging import set_up_default_logger

logger = logging.getLogger(__name__)


def constantinos_kalapotharakos_file_handle_to_sqlite(file_contents: bytes | mmap.mmap, output_file_path: Path,
                                                      parameter_count: int = 11):
    set_up_default_logger()
    output_file_path.unlink(missing_ok=True)
    Path(str(output_file_path) + '-shm').unlink(missing_ok=True)
    Path(str(output_file_path) + '-wal').unlink(missing_ok=True)
    Path(str(output_file_path) + '-journal').unlink(missing_ok=True)
    value_iterator = re.finditer(rb"[^\s]+", file_contents)
    phase_amplitude_count = 64
    parameter_column_names = [f'parameter{index}' for index in range(parameter_count)]
    phase_amplitude_column_names = [f'phase_amplitude{index}' for index in range(phase_amplitude_count)]
    data_column_names = parameter_column_names + phase_amplitude_column_names
    list_of_dictionaries: List[Dict] = []
    data_frame = pl.from_dicts([], schema={name: pl.Float32 for name in data_column_names})
    count = 0
    while True:
        parameters = []
        try:
            parameters.append(float(next(value_iterator).group(0)))
        except StopIteration:
            break
        for _ in range(parameter_count - 1):
            parameters.append(float(next(value_iterator).group(0)))
        _ = float(next(value_iterator).group(0))  # Likelihood in Constantinos' output which has no meaning here.
        phase_amplitudes = []
        for _ in range(phase_amplitude_count):
            phase_amplitudes.append(float(next(value_iterator).group(0)))
        row_values = parameters + phase_amplitudes
        row_dictionary = {name: value for name, value in zip(data_column_names, row_values)}
        list_of_dictionaries.append(row_dictionary)
        count += 1
        if len(list_of_dictionaries) % 100000 == 0:
            logger.info(f'Processed {count} rows.')
            chunk_data_frame = pl.from_dicts(list_of_dictionaries,
                                             schema={name: pl.Float32 for name in data_column_names})
            chunk_data_frame.write_database('main', f'sqlite:///{output_file_path}', if_table_exists='append')
            list_of_dictionaries = []
    chunk_data_frame = pl.from_dicts(list_of_dictionaries, schema={name: pl.Float32 for name in data_column_names})
    chunk_data_frame.write_database('main', f'sqlite:///{output_file_path}', if_table_exists='append')


def arbitrary_constantinos_kalapotharakos_file_path_to_pandas(data_path: Path, columns_per_row: int,
                                                              skip_rows: int = 0, limit: int | None = None
                                                              ) -> pd.DataFrame:
    polars_data_frame = arbitrary_constantinos_kalapotharakos_file_handle_to_polars(
        data_path=data_path, columns_per_row=columns_per_row, skip_rows=skip_rows, limit=limit)
    pandas_data_frame = polars_data_frame.to_pandas()
    return pandas_data_frame


def arbitrary_constantinos_kalapotharakos_file_handle_to_polars(data_path: Path, columns_per_row: int,
                                                                skip_rows: int = 0, limit: int | None = None
                                                                ) -> pl.DataFrame:
    with data_path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        return arbitrary_constantinos_kalapotharakos_file_contents_to_polars(
            file_contents, columns_per_row, skip_rows=skip_rows, limit=limit)


def combine_constantinos_kalapotharakos_split_output_files_to_csv(root_directory_path: Path, combined_output_path: Path,
                                                                  columns_per_row: int) -> None:
    split_data_frames: list[pl.DataFrame] = []
    for split_data_path in sorted(root_directory_path.glob('*.dat')):
        print(f'Processing {split_data_path}.')
        split_data_frame = arbitrary_constantinos_kalapotharakos_file_handle_to_polars(split_data_path,
                                                                                       columns_per_row=columns_per_row)
        rename_dictionary: dict[str, str] = {}
        for column_index in range(columns_per_row - 2):
            rename_dictionary[str(column_index)] = f'parameter{column_index}'
        rename_dictionary[str(columns_per_row - 2)] = f'log_likelihood'
        rename_dictionary[str(columns_per_row - 1)] = f'chain'
        split_data_frame = split_data_frame.rename(rename_dictionary)
        split_data_frame = split_data_frame.with_columns(split_data_frame["chain"].cast(pl.Int64).alias("chain"))
        split_data_frame_cpu_number = int(re.search('1(\d+)\.dat', split_data_path.name).group(1))
        split_data_frame = split_data_frame.with_columns(pl.lit(split_data_frame_cpu_number).alias('cpu'))
        iterations = math.ceil(split_data_frame.height / 2)
        iteration_array = np.arange(iterations, dtype=np.int64)
        combined_iteration_array = np.empty((iteration_array.size * 2), dtype=iteration_array.dtype)
        combined_iteration_array[0::2] = iteration_array
        combined_iteration_array[1::2] = iteration_array
        if split_data_frame.height % 2 != 0:
            combined_iteration_array = combined_iteration_array[:-1]
        split_data_frame = split_data_frame.with_columns(
            pl.Series(name='iteration', values=combined_iteration_array, dtype=pl.Int64))
        split_data_frames.append(split_data_frame)
    combined_data_frame = pl.concat(split_data_frames)
    combined_data_frame.write_csv(combined_output_path)


def arbitrary_constantinos_kalapotharakos_file_contents_to_polars(file_contents: bytes | mmap.mmap,
                                                                  columns_per_row: int, skip_rows: int = 0,
                                                                  limit: int | None = None) -> pl.DataFrame:
    set_up_default_logger()
    value_iterator = re.finditer(rb"[^\s]+", file_contents)
    list_of_dictionaries: List[Dict] = []
    data_frame = pl.from_dicts([], schema={str(index): pl.Float32 for index in range(columns_per_row)})
    count = 0
    while True:
        values = []
        try:
            values.append(float(next(value_iterator).group(0)))
        except StopIteration:
            break
        for _ in range(columns_per_row - 1):
            values.append(float(next(value_iterator).group(0)))
        if skip_rows > 0:
            skip_rows -= 1
            continue
        row_dictionary = {str(index): value for index, value in zip(range(columns_per_row), values)}
        list_of_dictionaries.append(row_dictionary)
        count += 1
        if limit is not None and count >= limit:
            break
        if len(list_of_dictionaries) % 100000 == 0:
            logger.info(f'Processed {count} rows.')
            chunk_data_frame = pl.from_dicts(list_of_dictionaries,
                                             schema={str(index): pl.Float32 for index in range(columns_per_row)})
            data_frame = data_frame.vstack(chunk_data_frame)
            list_of_dictionaries = []
    chunk_data_frame = pl.from_dicts(list_of_dictionaries,
                                     schema={str(index): pl.Float32 for index in range(columns_per_row)})
    data_frame = data_frame.vstack(chunk_data_frame)
    return data_frame


def get_memory_mapped_file_contents(file_handle: TextIO) -> mmap.mmap:
    file_fileno = file_handle.fileno()
    file_contents = mmap.mmap(file_fileno, 0, access=mmap.ACCESS_READ)
    return file_contents


def constantinos_kalapotharakos_format_file_to_sqlite(input_file_path: Path, output_file_path: Path,
                                                      parameter_count: int = 11) -> None:
    """
    Produces an SQLite database from a Constantinos Kalapotharakos format file. The expected input format includes
    11 parameters, 1 likelihood value (which is ignored), and 64 phase amplitude values for each entry.

    :param input_file_path: The Path to the Constantinos Kalapotharakos format file.
    :param output_file_path: The Path to the output SQLite database.
    """
    with input_file_path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        constantinos_kalapotharakos_file_handle_to_sqlite(file_contents, output_file_path, parameter_count)


if __name__ == '__main__':
    combine_constantinos_kalapotharakos_split_output_files_to_csv(
        Path('/Users/golmschenk/Downloads/mcmc_vac_all_5m_cont_with_phys'),
        Path('/Users/golmschenk/Downloads/mcmc_vac_all_5m_cont_with_phys.csv'),
        columns_per_row=13,
    )



from haplo.internal.constantinos_kalapotharakos_format import \
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr

combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr(,,



from __future__ import annotations

import itertools

import dask.array
import logging
import mmap
import numpy as np
import re
import shutil
import xarray
from pathlib import Path
from typing import Iterator, TextIO

logger = logging.getLogger(__name__)


class ConstantinosKalapotharakosFormatError(Exception):
    pass


def constantinos_kalapotharakos_format_record_generator(path: Path, elements_per_record: int
                                                        ) -> Iterator[tuple[float, ...]]:
    """
    Create a record generator for a Constantinos Kalapotharakos format file.

    :param path: The path to the file.
    :param elements_per_record: The number of elements per record.
    :return: A generator that iterates over the records.
    """
    with path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        generator = constantinos_kalapotharakos_format_record_generator_from_file_contents(file_contents=file_contents,
                                                                                           elements_per_record=elements_per_record)
        for record in generator:
            yield record


def constantinos_kalapotharakos_format_record_generator_from_file_contents(
        file_contents: bytes | mmap.mmap, elements_per_record: int) -> Iterator[tuple[float, ...]]:
    """
    Create a record generator for a Constantinos Kalapotharakos format file's contents.

    :param file_contents: The file contents object.
    :param elements_per_record: The number of elements per record.
    :return: A generator that iterates over the records.
    """
    value_iterator = re.finditer(rb"[^\s]+", file_contents)
    count = 0
    while True:
        values = []
        try:
            values.append(float(next(value_iterator).group(0)))
        except StopIteration:
            break
        try:
            for _ in range(elements_per_record - 1):
                values.append(float(next(value_iterator).group(0)))
            yield tuple(values)
            if count % 100000 == 0:
                logger.info(f'Processed {count} rows.')
            count += 1
        except StopIteration:
            raise ConstantinosKalapotharakosFormatError(
                f'The Constantinos Kalapotharakos format file ran out of elements when trying to get '
                f'{elements_per_record} elements for the current record.')


def get_memory_mapped_file_contents(file_handle: TextIO) -> mmap.mmap:
    """
    Get a memory mapped version of a file handle's contents.

    :param file_handle: The file handle to memory map.
    :return: The memory map.
    """
    file_fileno = file_handle.fileno()
    file_contents = mmap.mmap(file_fileno, 0, access=mmap.ACCESS_READ)
    return file_contents


def combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr(
        split_mcmc_output_directory: Path,
        combined_output_path: Path,
        *,
        elements_per_record: int,
        overwrite: bool = False
) -> None:
    """
    Combine Constantinos Kalapotharakos format split mcmc output files into an Xarray Zarr data store.

    :param split_mcmc_output_directory: The root of the split files.
    :param combined_output_path: The path of the output Zarr file.
    :param elements_per_record: The number of elements per record in the split files. Similar to columns per row, but
                                the files are not organized into rows and columns.
    :param overwrite: Overwrite existing files if they exist. Otherwise, an error will be raised if they exist.
    :return: None
    """

    def _save_batch_to_cpu_and_iteration_region(zarr_path, parameters_batch_, log_likelihood_batch_,
                                                region_start_iteration, region_end_iteration, cpu, chains_,
                                                parameter_count_, parameter_indexes_):
        flat_parameters_batch_array = np.array(parameters_batch_, dtype=np.float32)
        parameters_batch_array = flat_parameters_batch_array.reshape(
            [region_end_iteration - region_start_iteration, 1, chains_.size, parameter_count_])
        flat_log_likelihood_batch_array = np.array(log_likelihood_batch_, dtype=np.float32)
        log_likelihood_batch_array = flat_log_likelihood_batch_array.reshape(
            [region_end_iteration - region_start_iteration, 1, chains_.size])
        region_dataset = xarray.Dataset(
            coords={
                'iteration': np.arange(region_start_iteration, region_end_iteration, dtype=np.int64),
                'cpu': np.array([cpu], dtype=np.int64),
                'chain': chains_,
                'parameter_index': parameter_indexes_,
            },
            data_vars={
                'parameter': (
                    ['iteration', 'cpu', 'chain', 'parameter_index'],
                    parameters_batch_array,
                ),
                'log_likelihood': (
                    ['iteration', 'cpu', 'chain'],
                    log_likelihood_batch_array,
                ),
            },
        )
        region_dataset.to_zarr(zarr_path, region='auto')

    def _save_final_iteration_region(zarr_path, parameters_batch_, log_likelihood_batch_,
                                     iteration_, cpus_, chains_, parameter_count_, parameter_indexes_):
        flat_parameters_batch_array = np.array(parameters_batch_, dtype=np.float32)
        parameters_batch_array = flat_parameters_batch_array.reshape(
            [1, cpus_.size, chains_.size, parameter_count_])
        flat_log_likelihood_batch_array = np.array(log_likelihood_batch_, dtype=np.float32)
        log_likelihood_batch_array = flat_log_likelihood_batch_array.reshape(
            [1, cpus_.size, chains_.size])
        region_dataset = xarray.Dataset(
            coords={
                'iteration': np.array([iteration_], dtype=np.int64),
                'cpu': cpus_,
                'chain': chains_,
                'parameter_index': parameter_indexes_,
            },
            data_vars={
                'parameter': (
                    ['iteration', 'cpu', 'chain', 'parameter_index'],
                    parameters_batch_array,
                ),
                'log_likelihood': (
                    ['iteration', 'cpu', 'chain'],
                    log_likelihood_batch_array,
                ),
            },
        )
        region_dataset.to_zarr(zarr_path, append_dim='iteration')

    def _get_known_complete_iterations(split_data_file_paths, elements_per_record):
        split_data_path0 = split_data_file_paths[0]
        record_generator_ = constantinos_kalapotharakos_format_record_generator(
            split_data_path0, elements_per_record=elements_per_record)
        data_file0_iterations = 0
        is_final_iteration_known_incomplete_ = False
        read_chain0 = False
        for _ in record_generator_:
            if read_chain0:
                data_file0_iterations += 1
                read_chain0 = False
            else:
                read_chain0 = True
        if read_chain0:
            is_final_iteration_known_incomplete_ = True
        if is_final_iteration_known_incomplete_:
            max_known_complete_iteration_ = data_file0_iterations - 1
        else:
            max_known_complete_iteration_ = data_file0_iterations - 2
        return max_known_complete_iteration_, is_final_iteration_known_incomplete_

    def _create_empty_dataset_zarr(zarr_path, iterations, cpus, chains, parameter_indexes, iteration_chunk_size):
        empty_dataset = xarray.Dataset(
            coords={
                'iteration': iterations,
                'cpu': cpus,
                'chain': chains,
                'parameter_index': parameter_indexes,
            },
            data_vars={
                'parameter': (
                    ['iteration', 'cpu', 'chain', 'parameter_index'],
                    dask.array.full((iterations.size, cpus.size, chains.size, parameter_indexes.size),
                                    fill_value=np.nan, chunks=(iteration_chunk_size, -1, -1, -1), dtype=np.float32),
                ),
                'log_likelihood': (
                    ['iteration', 'cpu', 'chain'],
                    dask.array.full((iterations.size, cpus.size, chains.size),
                                    fill_value=np.nan, chunks=(iteration_chunk_size, -1, -1), dtype=np.float32),
                ),
            },
        )
        encoding = {
            'iteration': {'dtype': 'int64', 'chunks': (iteration_chunk_size,)},
            'cpu': {'dtype': 'int64', 'chunks': (-1,)},
            'chain': {'dtype': 'int64', 'chunks': (-1,)},
            'parameter_index': {'dtype': 'int64', 'chunks': (-1,)},
            'parameter': {'dtype': 'float32', 'chunks': (iteration_chunk_size, -1, -1, -1)},
            'log_likelihood': {'dtype': 'float32', 'chunks': (iteration_chunk_size, -1, -1)},
        }
        empty_dataset.to_zarr(zarr_path, compute=False, encoding=encoding)

    def _check_for_existing_files(combined_output_path_, overwrite_):
        temporary_combined_output_path_ = combined_output_path_.parent.joinpath(combined_output_path_.name + '.partial')
        if temporary_combined_output_path_.exists():
            if overwrite_:
                shutil.rmtree(temporary_combined_output_path_)
            else:
                raise FileExistsError(f'{temporary_combined_output_path_} needs to be created, but already exists. '
                                      f'Pass `overwrite=True` to overwrite.')
        if combined_output_path_.exists():
            if overwrite_:
                shutil.rmtree(combined_output_path_)
            else:
                raise FileExistsError(f'{combined_output_path_} needs to be created, but already exists. '
                                      f'Pass `overwrite=True` to overwrite.')
        return temporary_combined_output_path_

    temporary_combined_output_path = _check_for_existing_files(combined_output_path, overwrite)
    split_data_file_paths = sorted(split_mcmc_output_directory.glob('*.dat'))
    max_known_complete_iteration, is_final_iteration_known_incomplete = _get_known_complete_iterations(
        split_data_file_paths, elements_per_record)
    iterations = np.arange(max_known_complete_iteration + 1, dtype=np.int64)
    cpus = np.arange(len(split_data_file_paths), dtype=np.int64)
    chains = np.array([0, 1], dtype=np.int64)
    parameter_count = elements_per_record - 2
    parameter_indexes = np.arange(parameter_count, dtype=np.int64)
    iteration_chunk_size = 100
    _create_empty_dataset_zarr(temporary_combined_output_path, iterations, cpus, chains, parameter_indexes,
                               iteration_chunk_size)

    final_iteration_parameters_batch: list[tuple[float, ...]] = []
    final_iteration_log_likelihood_batch: list[float] = []
    for split_index, split_data_path in enumerate(split_data_file_paths):
        logger.info(f'Processing {split_data_path}.')
        record_generator = constantinos_kalapotharakos_format_record_generator(
            split_data_path, elements_per_record=elements_per_record)
        parameters_batch: list[tuple[float, ...]] = []
        log_likelihood_batch: list[float] = []
        split_data_frame_cpu_number = int(re.search(r'1(\d+)\.dat', split_data_path.name).group(1))
        if split_index != split_data_frame_cpu_number:
            raise ValueError(f'A split MCMC output file was expected but not found. '
                             f'Expected a file for CPU number {split_index}, but found {split_data_path.name}.')
        chain = 0
        iteration = 0
        batch_start_iteration = iteration
        for record_index in itertools.count():
            record = next(record_generator)
            if chain != int(record[elements_per_record - 1]):
                raise ValueError(f'The chain did not match the expected value at record index {record_index}.')
            parameters_batch.append(record[:parameter_count])
            log_likelihood_batch.append(record[parameter_count])
            if chain == 1:
                chain = 0
                iteration += 1
                if iteration > batch_start_iteration + iteration_chunk_size or iteration > max_known_complete_iteration:
                    _save_batch_to_cpu_and_iteration_region(temporary_combined_output_path, parameters_batch,
                                                            log_likelihood_batch, batch_start_iteration, iteration,
                                                            split_data_frame_cpu_number, chains, parameter_count,
                                                            parameter_indexes)
                    batch_start_iteration = iteration
                    parameters_batch = []
                    log_likelihood_batch = []
                if iteration > max_known_complete_iteration:
                    if not is_final_iteration_known_incomplete:
                        try:
                            record = next(record_generator)
                            final_iteration_parameters_batch.append(record[:parameter_count])
                            final_iteration_log_likelihood_batch.append(record[parameter_count])
                            record = next(record_generator)
                            final_iteration_parameters_batch.append(record[:parameter_count])
                            final_iteration_log_likelihood_batch.append(record[parameter_count])
                        except StopIteration:
                            is_final_iteration_known_incomplete = True
                    break
            else:
                chain = 1
    if not is_final_iteration_known_incomplete:
        _save_final_iteration_region(temporary_combined_output_path, final_iteration_parameters_batch,
                                     final_iteration_log_likelihood_batch,
                                     max_known_complete_iteration + 1, cpus, chains,
                                     parameter_count, parameter_indexes)
    temporary_combined_output_path.rename(combined_output_path)



import numpy as np
import pandas as pd
import pytest

from haplo.internal.mcmc_output_file_manipulations import get_last_states_of_mcmc_states_data_frame


mcmc_states_data_frame = pd.DataFrame(
    {
        'parameter': [0, 1, 2, 3, 4, 5, 6, 7, 8],
        'iteration': [0, 0, 0, 1, 1, 1, 2, 2, 2]
    }
)


def test_get_last_states_of_mcmc_data_frame():
    latest_mcmc_states_data_frame = get_last_states_of_mcmc_states_data_frame(mcmc_states_data_frame,
                                                                              number_of_states=5)
    expected_parameters = [4, 5, 6, 7, 8]
    assert np.allclose(np.sort(expected_parameters), np.sort(latest_mcmc_states_data_frame['parameter']))


def test_get_last_states_of_mcmc_data_frame_errors_when_requesting_larger_than_total():
    with pytest.raises(ValueError):
        _ = get_last_states_of_mcmc_states_data_frame(mcmc_states_data_frame, number_of_states=10)


def test_get_last_states_of_mcmc_data_frame_when_requested_is_smaller_than_last_iteration():
    latest_mcmc_states_data_frame = get_last_states_of_mcmc_states_data_frame(mcmc_states_data_frame,
                                                                              number_of_states=2)
    expected_parameters = [7, 8]
    assert np.allclose(np.sort(expected_parameters), np.sort(latest_mcmc_states_data_frame['parameter']))



import datetime
import sys
from pathlib import Path

sys.path.insert(0, str(Path("../../src/haplo").absolute()))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "haplo"
project_copyright = f'2023 - {datetime.datetime.now().year}, Greg Olmschenk'
author = "Greg Olmschenk"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
]

templates_path = ["_templates"]
exclude_patterns = []
source_suffix = [".rst", ".md"]
autodoc_class_signature = 'separated'
autodoc_default_options = {
    'special-members': None,
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



import numpy as np
from pathlib import Path

from haplo.nicer_dataset import NicerDataset, split_dataset_into_count_datasets
from haplo.nicer_transform import PrecomputedNormalizeParameters, PrecomputedNormalizePhaseAmplitudes


def example_infer_session():
    print('0')
    full_dataset_path = Path('data/640m_parameters_and_phase_amplitudes.db')
    full_train_dataset = NicerDataset.new(
        dataset_path=full_dataset_path,
        # parameters_transform=PrecomputedNormalizeParameters(),
        # phase_amplitudes_transform=PrecomputedNormalizePhaseAmplitudes(),
        length=200_000,
    )
    test_dataset, validation_dataset, train_dataset, _ = split_dataset_into_count_datasets(
        full_train_dataset, [100_000, 10, 10])
    print('1')

    parameters_list = []
    test_phase_amplitudes_list = []
    for test_parameters, test_phase_amplitudes in test_dataset:
        parameters_list.append(test_parameters)
        test_phase_amplitudes_list.append(test_phase_amplitudes)
        if len(parameters_list) % 1000 == 0:
            print(len(parameters_list))
    parameters_array = np.stack(parameters_list, axis=0, dtype=np.float32)
    phase_amplitudes_array = np.stack(test_phase_amplitudes_list, axis=0, dtype=np.float32)
    np.save(Path('test_parameters.npy'), parameters_array)
    np.save(Path('test_phase_amplitudes.npy'), phase_amplitudes_array)

if __name__ == '__main__':
    example_infer_session()



import numpy as np
import pytest
import xarray as xr

from haplo.internal.mcmc_output_analysis import slice_iteration_of_mcmc_output_xarray_dataset, \
    mcmc_output_xarray_dataset_to_pandas_data_frame


@pytest.fixture
def sample_dataset():
    number_of_iterations = 10
    number_of_cpus = 4
    number_of_chains = 2
    number_of_parameters = 3

    iterations = np.arange(number_of_iterations)
    cpus = np.arange(number_of_cpus)
    chains = np.arange(number_of_chains)
    parameters = np.arange(number_of_parameters)

    parameter_data = np.random.rand(number_of_iterations, number_of_cpus, number_of_chains, number_of_parameters)
    log_likelihood_data = np.random.rand(number_of_iterations, number_of_cpus, number_of_chains)

    dataset = xr.Dataset({
        'parameter': (['iteration', 'cpu', 'chain', 'parameter_index'], parameter_data),
        'log_likelihood': (['iteration', 'cpu', 'chain'], log_likelihood_data)
    }, coords={
        'iteration': iterations,
        'cpu': cpus,
        'chain': chains,
        'parameter_index': parameters
    })

    return dataset


def test_slice_iteration_of_mcmc_output_xarray_dataset(sample_dataset):
    start_iteration = 2
    end_iteration = 5

    sliced_dataset = slice_iteration_of_mcmc_output_xarray_dataset(
        sample_dataset, start_iteration, end_iteration
    )

    assert len(sliced_dataset['iteration']) == end_iteration - start_iteration
    assert all(value in range(start_iteration, end_iteration) for value in sliced_dataset['iteration'].values)


def test_mcmc_output_xarray_dataset_to_pandas_data_frame(sample_dataset):
    data_frame = mcmc_output_xarray_dataset_to_pandas_data_frame(sample_dataset)

    parameter_columns = [f'parameter{parameter_index}' for parameter_index in sample_dataset['parameter_index'].values]
    expected_columns = parameter_columns + ['log_likelihood']
    assert list(data_frame.columns) == expected_columns

    expected_rows = sample_dataset['iteration'].size * sample_dataset['cpu'].size * sample_dataset['chain'].size
    assert data_frame.shape[0] == expected_rows

    assert (data_frame.loc[5, 2, 1]['parameter1'] ==
            sample_dataset.sel({'iteration': 5, 'cpu': 2, 'chain': 1})['parameter'][1].item())


def test_mcmc_output_xarray_dataset_to_pandas_data_frame_with_limit_from_end(sample_dataset):
    data_frame = mcmc_output_xarray_dataset_to_pandas_data_frame(sample_dataset, limit_from_end=10)

    assert data_frame.shape[0] == 10


def test_mcmc_output_xarray_dataset_to_pandas_data_frame_with_sample_size(sample_dataset):
    data_frame = mcmc_output_xarray_dataset_to_pandas_data_frame(sample_dataset, random_sample_size=10)

    assert data_frame.shape[0] == 10



import pytest
import xarray
from pathlib import Path

import shutil

from haplo.internal.constantinos_kalapotharakos_format import \
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr


def test_combine_constantinos_kalapotharakos_split_mcmc_output():
    root_split_files_directory = Path(__file__).parent.joinpath(
        'combine_constantinos_kalapotharakos_split_mcmc_output_resources')
    output_path = root_split_files_directory.joinpath('output.zarr')
    if output_path.exists():
        shutil.rmtree(output_path)
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr(root_split_files_directory, output_path,
                                                                               elements_per_record=13, overwrite=True)
    xarray_dataset = xarray.load_dataset(output_path)
    assert xarray_dataset['parameter'].shape == (3, 4, 2, 11)
    assert xarray_dataset['iteration'].max() == 2
    assert xarray_dataset['parameter'][1, 1, 0, 8].item() == pytest.approx(2.06011819056089)
    assert xarray_dataset['log_likelihood'][2, 3, 1].item() == pytest.approx(-24990.2909981251)
    assert xarray_dataset['parameter'].encoding['chunks'] == (100, 4, 2, 11)
    if output_path.exists():
        shutil.rmtree(output_path)


def test_combine_constantinos_kalapotharakos_split_mcmc_output_with_complete_final_iteration():
    root_split_files_directory = Path(__file__).parent.joinpath(
        'combine_constantinos_kalapotharakos_split_mcmc_output_resources_with_complete_final_iteration')
    output_path = root_split_files_directory.joinpath('output.zarr')
    if output_path.exists():
        shutil.rmtree(output_path)
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr(root_split_files_directory, output_path,
                                                                               elements_per_record=13, overwrite=True)
    xarray_dataset = xarray.load_dataset(output_path)
    assert xarray_dataset['parameter'].shape == (4, 4, 2, 11)
    assert xarray_dataset['iteration'].max() == 3
    assert xarray_dataset['parameter'][1, 1, 0, 8].item() == pytest.approx(2.06011819056089)
    assert xarray_dataset['log_likelihood'][2, 3, 1].item() == pytest.approx(-24990.2909981251)
    assert xarray_dataset['parameter'].encoding['chunks'] == (100, 4, 2, 11)
    if output_path.exists():
        shutil.rmtree(output_path)


def test_combine_constantinos_kalapotharakos_split_mcmc_output_with_incomplete_final_iteration():
    root_split_files_directory = Path(__file__).parent.joinpath(
        'combine_constantinos_kalapotharakos_split_mcmc_output_resources_with_incomplete_final_iteration')
    output_path = root_split_files_directory.joinpath('output.zarr')
    if output_path.exists():
        shutil.rmtree(output_path)
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr(root_split_files_directory, output_path,
                                                                               elements_per_record=13, overwrite=True)
    xarray_dataset = xarray.load_dataset(output_path)
    assert xarray_dataset['parameter'].shape == (3, 4, 2, 11)
    assert xarray_dataset['iteration'].max() == 2
    assert xarray_dataset['parameter'][1, 1, 0, 8].item() == pytest.approx(2.06011819056089)
    assert xarray_dataset['log_likelihood'][2, 3, 1].item() == pytest.approx(-24990.2909981251)
    assert xarray_dataset['parameter'].encoding['chunks'] == (100, 4, 2, 11)
    if output_path.exists():
        shutil.rmtree(output_path)



from __future__ import annotations

import pandas as pd
from numpy.random import RandomState


def get_last_states_of_mcmc_states_data_frame(mcmc_states_data_frame: pd.DataFrame, number_of_states: int,
                                              iteration_column_name: str = 'iteration',
                                              maximum_iteration_to_include: int | None = None
                                              ) -> pd.DataFrame:
    random_state = RandomState(0)
    minimum_iteration = mcmc_states_data_frame[iteration_column_name].min()
    if maximum_iteration_to_include is None:
        maximum_iteration = mcmc_states_data_frame[iteration_column_name].max()
    else:
        maximum_iteration = maximum_iteration_to_include
    current_iteration = maximum_iteration
    latest_states_data_frame: pd.DataFrame = mcmc_states_data_frame[
        mcmc_states_data_frame[iteration_column_name] == current_iteration]
    if latest_states_data_frame.shape[0] >= number_of_states:
        latest_states_data_frame = latest_states_data_frame.sample(number_of_states, random_state=random_state)
        latest_states_data_frame.sort_values(iteration_column_name)
        return latest_states_data_frame
    while current_iteration > minimum_iteration:
        current_iteration -= 1
        current_iteration_states_data_frame: pd.DataFrame = mcmc_states_data_frame[
            mcmc_states_data_frame[iteration_column_name] == current_iteration]
        if latest_states_data_frame.shape[0] + current_iteration_states_data_frame.shape[0] >= number_of_states:
            current_iteration_states_data_frame = current_iteration_states_data_frame.sample(
                number_of_states - latest_states_data_frame.shape[0], random_state=random_state)
            latest_states_data_frame = pd.concat([current_iteration_states_data_frame, latest_states_data_frame])
            latest_states_data_frame = latest_states_data_frame.sort_values(iteration_column_name)
            return latest_states_data_frame
        else:
            latest_states_data_frame = pd.concat([current_iteration_states_data_frame, latest_states_data_frame])
    raise ValueError(f'Exhausted MCMC state data frame while attempting to obtain f{number_of_states}.')



from __future__ import annotations

import pandas as pd
import xarray
from numpy.random import RandomState
from pathlib import Path

from haplo.internal.constantinos_kalapotharakos_format import \
    combine_constantinos_kalapotharakos_split_mcmc_output_files_to_xarray_zarr


def slice_iteration_of_mcmc_output_xarray_dataset(
        dataset: xarray.Dataset,
        *,
        start_iteration: int,
        end_iteration: int,
) -> xarray.Dataset:
    """
    Gets a slice of an MCMC output Xarray dataset along the iteration axis.

    :param dataset: The MCMC output Xarray dataset.
    :param start_iteration: The start of the slice (inclusive).
    :param end_iteration: The end of the slice (exclusive).
    :return: The Xarray dataset that is the slice of the original dataset.
    """
    sliced_dataset = dataset.sel({'iteration': slice(start_iteration, end_iteration - 1)})
    return sliced_dataset


def mcmc_output_xarray_dataset_to_pandas_data_frame(
        dataset: xarray.Dataset,
        limit_from_end: int | None = None,
        random_sample_size: int | None = None,
) -> pd.DataFrame:
    """
    Converts the MCMC output Xarray dataset to a Pandas data frame.

    :param dataset: The MCMC output Xarray dataset.
    :param limit_from_end: Limits the number of rows from the end of the dataset.
    :param random_sample_size: Randomly samples this many elements from the dataset. If `limit_from_end` is set, that
                               limit will be applied first, then this many will be sampled from that subset.
    :return: The Pandas data frame.
    """
    state_parameter_data_array = dataset['parameter'].stack({'state_index': ['iteration', 'cpu', 'chain']}).transpose()
    state_log_likelihood_data_array = dataset['log_likelihood'].stack({'state_index': ['iteration', 'cpu', 'chain']}
                                                                      ).transpose()
    if limit_from_end is not None:
        state_parameter_data_array = state_parameter_data_array.isel(
            {'state_index': slice(-limit_from_end, None)})
        state_log_likelihood_data_array = state_log_likelihood_data_array.isel(
            {'state_index': slice(-limit_from_end, None)})
    if random_sample_size is not None:
        random_state = RandomState(0)
        sample_indexes = random_state.choice(state_parameter_data_array['state_index'].size, random_sample_size,
                                             replace=False)
        state_parameter_data_array = state_parameter_data_array.isel({'state_index': sample_indexes})
        state_log_likelihood_data_array = state_log_likelihood_data_array.isel({'state_index': sample_indexes})
    parameter_data_frame = state_parameter_data_array.to_pandas()
    parameter_data_frame.rename(columns={parameter_index: f'parameter{parameter_index}'
                                         for parameter_index in parameter_data_frame.columns}, inplace=True)
    log_likelihood_data_frame = state_log_likelihood_data_array.to_pandas()
    data_frame = pd.concat([parameter_data_frame, log_likelihood_data_frame], axis=1)
    if random_sample_size is not None:
        data_frame.sort_index(inplace=True)
    return data_frame



from __future__ import annotations

from enum import StrEnum
import itertools
import logging
import math
import mmap
import numpy as np
import pandas as pd
import re
import shutil
import xarray
import zarr
from pandas import DataFrame
from pathlib import Path
from xarray import Dataset

from haplo.data_preparation import get_memory_mapped_file_contents, \
    arbitrary_constantinos_kalapotharakos_file_path_to_pandas
from haplo.logging import set_up_default_logger

logger = logging.getLogger(__name__)


class DatasetVariableName(StrEnum):
    INPUT = 'input'
    OUTPUT = 'output'


def constantinos_kalapotharakos_format_file_to_xarray_zarr(
        input_path: Path,
        output_path: Path,
        input_size: int = 11,
        output_size: int = 64,
        zarr_chunk_axis0_size: int = 1000,
) -> None:
    set_up_default_logger()
    if output_path.exists():
        shutil.rmtree(output_path)
    with input_path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        value_iterator = re.finditer(rb"[^\s]+", file_contents)
        parameters_set = []
        phase_amplitudes_set = []
        encoding = {
            DatasetVariableName.INPUT: {'dtype': 'float32', 'chunks': (zarr_chunk_axis0_size, input_size)},
            DatasetVariableName.OUTPUT: {'dtype': 'float32', 'chunks': (zarr_chunk_axis0_size, output_size)},
        }
        for index in itertools.count():
            parameters = []
            try:
                parameters.append(float(next(value_iterator).group(0)))
            except StopIteration:
                break
            for _ in range(input_size - 1):
                parameters.append(float(next(value_iterator).group(0)))
            _ = float(next(value_iterator).group(0))  # Likelihood in Constantinos' output which has no meaning here.
            phase_amplitudes = []
            for _ in range(output_size):
                phase_amplitudes.append(float(next(value_iterator).group(0)))
            parameters_set.append(parameters)
            phase_amplitudes_set.append(phase_amplitudes)
            if (index + 1) % 100000 == 0:
                partial_dataset = xarray.Dataset(data_vars={
                    DatasetVariableName.INPUT: (['index', 'parameter_index'], parameters_set),
                    DatasetVariableName.OUTPUT: (['index', 'phase_index'], phase_amplitudes_set),
                })
                if not output_path.exists():
                    partial_dataset.to_zarr(output_path, encoding=encoding)
                else:
                    partial_dataset.to_zarr(output_path, append_dim='index')
                logger.info(f'Processed {index + 1} rows.')
                parameters_set = []
                phase_amplitudes_set = []
        if len(parameters_set) != 0:
            partial_dataset = xarray.Dataset(data_vars={
                DatasetVariableName.INPUT: (['index', 'parameter_index'], parameters_set),
                DatasetVariableName.OUTPUT: (['index', 'phase_index'], phase_amplitudes_set),
            })
            if not output_path.exists():
                partial_dataset.to_zarr(output_path, encoding=encoding)
            else:
                partial_dataset.to_zarr(output_path, append_dim='index')


def convert_directory_xarray_zarr_to_zip_xarray_zarr(
        input_path: Path,
        output_path: Path,
) -> None:
    if output_path.suffix != '.zip':
        raise ValueError(f'Expected a .zip extension for the output file {output_path}')
    dataset = xarray.open_zarr(input_path)
    dataset.to_zarr(output_path, mode='w')


def constantinos_kalapotharakos_format_file_to_xarray_zarr_zip(
        input_path: Path,
        output_path: Path,
        input_size: int = 11,
        output_size: int = 64,
        zarr_chunk_axis0_size: int = 1000,
) -> None:
    if output_path.suffix != '.zip':
        raise ValueError(f'Expected a .zip extension for the output file {output_path}')
    temporary_intermediate_unzipped_zarr_path = output_path.parent.joinpath(output_path.stem + '.zarr')
    if temporary_intermediate_unzipped_zarr_path.exists():
        raise ValueError(f'Tried to use temporary file {temporary_intermediate_unzipped_zarr_path}, but it already '
                         f'exists.')
    constantinos_kalapotharakos_format_file_to_xarray_zarr(
        input_path=input_path,
        output_path=temporary_intermediate_unzipped_zarr_path,
        input_size=input_size,
        output_size=output_size,
        zarr_chunk_axis0_size=zarr_chunk_axis0_size,
    )
    convert_directory_xarray_zarr_to_zip_xarray_zarr(
        input_path=temporary_intermediate_unzipped_zarr_path, output_path=output_path)
    shutil.rmtree(temporary_intermediate_unzipped_zarr_path)


def constantinos_kalapotharakos_format_file_to_zarr(input_file_path: Path, output_file_path: Path,
                                                    parameter_count: int = 11) -> None:
    """
    Converts a Constantinos Kalapotharakos format text file to a Zarr data store.

    :param input_file_path: The path to the text file.
    :param output_file_path: The path of the data store.
    :param parameter_count: The number of parameters per row.
    :return: None
    """
    with input_file_path.open() as file_handle:
        file_contents = get_memory_mapped_file_contents(file_handle)
        constantinos_kalapotharakos_file_handle_to_1d_input_1d_output_zarr(file_contents, output_file_path,
                                                                           parameter_count)


def constantinos_kalapotharakos_file_handle_to_1d_input_1d_output_zarr(file_contents: bytes | mmap.mmap,
                                                                       output_file_path: Path,
                                                                       input_size: int = 11,
                                                                       output_size: int = 64,
                                                                       zarr_chunk_axis0_size: int = 1000) -> None:
    """
    Converts the file contents of a Constantinos Kalapotharakos format text file to a Zarr data store.

    :param file_contents: Data to parse, provided in binary form as either bytes or mmap.
    :param output_file_path: Path where the resulting Zarr dataset is stored.
    :param input_size: Number of elements expected in the input array per data row.
    :param output_size: Number of elements expected in the output array per data row.
    :param zarr_chunk_axis0_size: Number of rows to store per Zarr chunk (axis 0).
    :return: None
    """
    set_up_default_logger()
    zarr_store = zarr.DirectoryStore(str(output_file_path))
    root = zarr.group(store=zarr_store, overwrite=True)
    input_array = root.create_dataset(
        DatasetVariableName.INPUT,
        shape=(0, input_size),
        chunks=(zarr_chunk_axis0_size, input_size),
        dtype='float32',
    )
    output_array = root.create_dataset(
        DatasetVariableName.OUTPUT,
        shape=(0, output_size),
        chunks=(zarr_chunk_axis0_size, output_size),
        dtype='float32',
    )

    value_iterator = re.finditer(rb"[^\s]+", file_contents)
    parameters_set = []
    phase_amplitudes_set = []
    for index in itertools.count():
        parameters = []
        try:
            parameters.append(float(next(value_iterator).group(0)))
        except StopIteration:
            break
        for _ in range(input_size - 1):
            parameters.append(float(next(value_iterator).group(0)))
        _ = float(next(value_iterator).group(0))  # Likelihood in Constantinos' output which has no meaning here.
        phase_amplitudes = []
        for _ in range(output_size):
            phase_amplitudes.append(float(next(value_iterator).group(0)))
        parameters_set.append(parameters)
        phase_amplitudes_set.append(phase_amplitudes)
        if (index + 1) % 100000 == 0:
            input_array.append(parameters_set)
            output_array.append(phase_amplitudes_set)
            logger.info(f'Processed {index + 1} rows.')
            parameters_set = []
            phase_amplitudes_set = []
    input_array.append(parameters_set)
    output_array.append(phase_amplitudes_set)


def persist_xarray_dataset_variable_order(dataset: Dataset) -> Dataset:
    return dataset.assign_attrs(variable_order=[variable_name for variable_name in dataset.variables
                                                if variable_name not in dataset.dims])


def to_ordered_dataframe(dataset: Dataset) -> DataFrame:
    variable_order = dataset.attrs['variable_order']
    data_frame = dataset.to_dataframe()
    data_frame = data_frame[variable_order]
    return data_frame


def convert_from_2d_xarray_zarr_to_csv(xarray_zarr_path: Path, csv_path: Path) -> None:
    """
    Convert a table-like 2D Xarray Zarr data store to a CSV file.

    :param xarray_zarr_path: The path to the Zarr data store.
    :param csv_path: The path to the CSV file.
    :return: None
    """
    dataset = xarray.open_zarr(xarray_zarr_path)
    data_frame: pd.DataFrame = to_ordered_dataframe(dataset)
    data_frame.to_csv(csv_path, index=False)
